;; Generated package description from popper.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "popper" "0.4.6.0.20250215.81351" "Summon and dismiss buffers as popups" '((emacs "26.1")) :commit "c0dced3cdff15d3ebfc56a52d7616b390255eb76" :authors '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")) :maintainer '("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com") :keywords '("convenience") :url "https://github.com/karthink/popper")
